#ifndef F_CPU
	#define F_CPU 8000000UL
#endif
#include <stdlib.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/eeprom.h>

/* 9600 baud */
#define BAUDRATE 9600UL

uint8_t EEMEM numID;
uint8_t EEMEM knownID[10][12];

void uart_init( void ) {
//    uint16_t ubrr = (uint16_t) ((uint32_t) F_CPU/(16*BAUDRATE) - 1);
//	UBRRL  = ubrr;      //Baudrate 38400
	UBRRL  = 51;
	UCSRA |= 1;      //Doppelte Baudrate
	UCSRB  = 0x18;      //RX und TX erlaubt

    // Flush Receive-Buffer (entfernen evtl. vorhandener ungültiger Werte) 
    do {
        UDR;
    } while( UCSRA & (1 << RXC) );
}
static inline int uart_putc( const uint8_t c ) {
    // Warten, bis UDR bereit ist für einen neuen Wert 
    while( !( UCSRA & (1 << UDRE) ) )
        ;

    // UDR Schreiben startet die Übertragung 
    UDR = c;

    return( 1 );
}

static inline uint8_t uart_getc_wait( void ) {
    // Warten, bis etwas empfangen wird 
    while( !( UCSRA & (1 << RXC) ) )
        ;

    // Das empfangene Zeichen zurückliefern 
    return( UDR );
}

static inline int uart_getc_nowait( void ) {
    // Liefer das empfangene Zeichen, falls etwas empfangen wurde; -1 sonst 
    return( UCSRA & (1 << RXC) ) ? (int)UDR : -1;
}

void uart_puts( const char *data ) {
	/* send string per UART
	*/
	while( *data != 0 ) {
		uart_putc( *data++ );
	}
}

void IOinit( void ) {
	// PD2, PD3 und PD4 als Ausgang
	// PD2: Relais
	// PD3: Gruene LED
	// PD4: Rote LED
	DDRD |= ( 1<<PD2 ) | ( 1<<PD3 ) | ( 1 <<PD4 );

	// Alle Ausgänge ausschalten
	PORTD &= ~( 1<<PD2 ) & ~( 1<<PD3 ) & ~( 1<<PD4 );
}

void pause( uint16_t ms ) {
	/* pauses the program for the given time;
	* ms: time in milliseconds (0 - 65535)
	*/

	uint16_t i;

	for( i = 0; i < ms / 63; i++ ) {
		_delay_ms( 63 );
	}

	_delay_ms( ms % 63 );
}

uint8_t knownCard( const char *ID ) {
	uint8_t i,j,res;
	char eeID[ 12 ];

	eeprom_busy_wait();
	// Codes der Reihe nach durchgehen

	for( i = 0; i <= eeprom_read_byte( &numID ); i++ ) {
		eeprom_read_block( (void*)&eeID, (const void*)&knownID[ i ], 12 );
		res = 0;
		for( j = 0; j < 12; j++ ) {
			if( eeID[ j ] != ID[ j ] ) {
				res = 1;
				break;
			}
		}
		if( res == 0 ) {
			break;
		}
	}
	return( res );
}

int main( void ) {
	IOinit();
    uart_init(); 
	char *endln  = ".\r\n";

	char cardID[ 12 ];
	char tempID;

	uint8_t cardcnt = 0;
	uint8_t eecnt = 0;

	uart_puts( "Ready\r\n" );
	while( 1 ) {
		switch( uart_getc_wait() ) {
			case 0x02:					// bei 0x02 beginnt Code
				for( cardcnt = 0; cardcnt < 12; cardcnt++ ) {
					tempID = uart_getc_wait();
					cardID[ cardcnt ] = tempID;
				}
				break;
			case 'W':				// W für write
				eeprom_busy_wait();
				eecnt = eeprom_read_byte( &numID );
				if( eecnt == 9 ) {	// Wenn bereits 10 werte gespeichert sind passt nix mehr rein
					uart_puts( "EEF" );
					continue;
				} else {
					for( cardcnt = 0; cardcnt < 12; cardcnt++ ) {
						// Der Reihe nach die Code-Digits einlesen
						tempID = uart_getc_wait();
						cardID[ cardcnt ] = tempID;
					}
					// Dann in EEPROM an Position + 1 speichern
					eeprom_update_block( (const void*)&cardID, (void*)&knownID[ eecnt + 1 ][0], 12 );
					// Zum Schluss neue Anzahl von Codes in EEPROM speichern
					eeprom_update_byte( (void*)&numID, ( eecnt + 1 ) );
					cardcnt = 0;
				}
				pause( 50 );
				break;
			case 'R':				// R für read
				eeprom_busy_wait();
				for( cardcnt = 0; cardcnt <= eeprom_read_byte( &numID ); cardcnt++ ) {
					//uart_putc( 48 + cardcnt );		// macht Programm  ebenfalls zu groß :(
					//uart_putc( ':' );
					eeprom_read_block( (void*)&cardID, (const void*)&knownID[ cardcnt ], 12 );
					uart_puts( cardID );
					uart_puts( endln );
				}
				cardcnt = 0;
				break;
			case 'X':				// X für delete
				eecnt = 0;
				while( eecnt < 48 || eecnt > 58 ) {	// nur Zahlen einlesen
					eecnt = uart_getc_wait();
				}
				eecnt = eecnt - 48;	// ASCII-Zahl zu Zahl "konvertieren"
				eeprom_busy_wait();
				if( eecnt > eeprom_read_byte( &numID ) ) {
					uart_puts( "BID" );
				} else {
					for( cardcnt = eecnt; cardcnt < 9; cardcnt++ ) {
						// IDs in 12er bloecken nach vorne schieben....
						eeprom_read_block( (void*)&cardID, (const void*)&knownID[ cardcnt + 1 ], 12 );
						eeprom_busy_wait();
						eeprom_update_block( (const void*)&cardID, (void*)&knownID[ cardcnt ][0], 12 );
					}
					// Anzahl von Codes um 1 weniger
					eecnt = eeprom_read_byte( &numID );
					if( eecnt > 0 ) {
						eeprom_busy_wait();
						eeprom_update_byte( (void*)&numID, ( eecnt - 1 ) );
					}
					uart_puts( "DID" );
				}
				cardcnt = 0;
				break;

			case 'L':			// L für erase
				eeprom_busy_wait();
				eeprom_update_byte( (void*)&numID, 0x00 );
				uart_puts( "EID" );
				break;

			default:
				cardcnt = 0;
		}

		uart_puts( "\r\n" );

		if( cardcnt != 12 ) {
			cardcnt = 0;
			pause( 50 );
			continue;
		}

		if( knownCard( cardID ) == 0 ) {
			uart_puts( "OK: " );
			uart_puts( cardID );
			uart_puts( endln );
			PORTD |= ( 1<<PD2 ) | ( 1<<PD3 );	// Gruene LED + Relais ein
			pause( 2000 ); // 2 sekunden pause machen
			PORTD &= ~( 1<<PD2 ) & ~( 1<<PD3 );	// Gruene LED + Relais aus
		} else {
			uart_puts( "NOK: " );
			uart_puts( cardID );
			uart_puts( endln );

			for( cardcnt = 0; cardcnt < 4; cardcnt++ ) {
				PORTD |= ( 1<<PD4 );	// rote LED blinken...
				pause( 250 );
				PORTD &= ~( 1<<PD4 );
				pause( 250 );
			}
		}

		cardcnt = 0;
	}

	return( 0 );
}
